"""The core implementation of Inception Score and FID."""

from typing import List, Union, Tuple

import numpy as np
import torch
from scipy import linalg
from tqdm.auto import tqdm
from torch.utils.data import DataLoader, SequentialSampler, TensorDataset
from torch.utils.data.distributed import DistributedSampler

from . import districuted as dist
from .inception import InceptionV3


# device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')


def get_inception_feature(
    images: Union[torch.FloatTensor, DataLoader],
    dims: List[int],
    batch_size: int = 50,
    use_torch: bool = False,
    verbose: bool = False,
    device: torch.device = None,
) -> Union[torch.FloatTensor, np.ndarray]:
    """Calculate Inception Score and FID.

    For each image, only a forward propagation is required to calculating
    features for FID and Inception Score.

    Args:
        images: tensor or torch.utils.data.Dataloader. The images
                must be float tensor of range [0, 1].
        dims: List of int, see InceptionV3.BLOCK_INDEX_BY_DIM for
              available dimension.
        batch_size: int, The batch size for calculating activations. If
                    `images` is torch.utils.data.Dataloader, this argument is
                    ignored.
        use_torch: When True, use torch to calculate FID. Otherwise, use numpy.
        verbose: Set verbose to False for disabling progress bar. Otherwise,
                 the progress bar is showing when calculating activations.
        device: the torch device which is used to calculate inception feature
    Returns:
        inception_features: a list of extracted inception features
            corresponding to given dims.
    """
    assert all(dim in InceptionV3.BLOCK_INDEX_BY_DIM for dim in dims)
    if device is None:
        device = dist.device()

    if not isinstance(images, DataLoader):
        num_images = len(images)
        if dist.world_size() > 1:
            sampler = DistributedSampler(
                TensorDataset(images), shuffle=False)
        else:
            sampler = SequentialSampler(TensorDataset(images))
        # print(sampler)
        loader = DataLoader(
            images,
            batch_size=batch_size,
            sampler=sampler,
            drop_last=False,
        )
    else:
        num_images = len(images.dataset)
        loader = images

    block_idxs = [InceptionV3.BLOCK_INDEX_BY_DIM[dim] for dim in dims]
    # Initialize InceptionV3 model
    if dist.rank() == 0:
        # Only rank 0 initialize download the model
        model = InceptionV3(block_idxs).to(device)
        dist.barrier()
    else:
        # Other ranks wait until rank 0 download the model
        dist.barrier()
        model = InceptionV3(block_idxs).to(device)
    model.eval()

    if dist.rank() == 0:
        if use_torch:
            features = [
                torch.empty((num_images, dim)).to(device)
                for dim in dims]
        else:
            features = [
                np.empty((num_images, dim))
                for dim in dims]

    pbar = tqdm(
        total=num_images, dynamic_ncols=True, leave=False,
        disable=(dist.rank() != 0 or not verbose),
        desc="get_inception_feature")
    start = 0
    for batch_images in loader:
        batch_images = batch_images.to(device)
        # calculate inception feature
        end = min(start + len(batch_images) * dist.world_size(), num_images)
        with torch.no_grad():
            outputs = model(batch_images)
            if end == num_images:
                # This is the last batch, so we need to remove the padding.
                if num_images % dist.world_size() != 0:
                    is_padded = dist.rank() >= num_images % dist.world_size()
                else:
                    is_padded = False
                if is_padded:
                    # Remove the padding
                    outputs = [output[: -1] for output in outputs]
                    for output in outputs:
                        print(output.shape)
            outputs = [dist.gather(output) for output in outputs]
            if dist.rank() == 0:
                for feature, output, dim in zip(features, outputs, dims):
                    if use_torch:
                        feature[start: end] = output.view(-1, dim)
                    else:
                        feature[start: end] = output.view(-1, dim).cpu().numpy()
                dist.barrier()
            else:
                dist.barrier()
        pbar.update(end - start)
        start = end
    pbar.close()
    assert start == num_images
    if dist.rank() == 0:
        return features
    else:
        return [None for _ in range(len(dims))]


def torch_cov(m, rowvar=False):
    """Estimate a covariance matrix given data.

    Covariance indicates the level to which two variables vary together.
    If we examine N-dimensional samples, `X = [x_1, x_2, ... x_N]^T`,
    then the covariance matrix element `C_{ij}` is the covariance of
    `x_i` and `x_j`. The element `C_{ii}` is the variance of `x_i`.

    Args:
        m: A 1-D or 2-D array containing multiple variables and observations.
           Each row of `m` represents a variable, and each column a single
           observation of all those variables.
        rowvar: If `rowvar` is True, then each row represents a
                variable, with observations in the columns. Otherwise, the
                relationship is transposed: each column represents a variable,
                while the rows contain observations.

    Returns:
        The covariance matrix of the variables.
    """
    if m.dim() > 2:
        raise ValueError('m has more than 2 dimensions')
    if m.dim() < 2:
        m = m.view(1, -1)
    if not rowvar and m.size(0) != 1:
        m = m.t()
    # m = m.type(torch.double)  # uncomment this line if desired
    fact = 1.0 / (m.size(1) - 1)
    m -= torch.mean(m, dim=1, keepdim=True)
    mt = m.t()  # if complex: mt = m.t().conj()
    return fact * m.matmul(mt).squeeze()


# Pytorch implementation of matrix sqrt, from Tsung-Yu Lin, and Subhransu Maji
# https://github.com/msubhransu/matrix-sqrt
def sqrt_newton_schulz(A, numIters, dtype=None): # noqa
    with torch.no_grad():
        if dtype is None:
            dtype = A.type()
        batchSize = A.shape[0]
        dim = A.shape[1]
        normA = A.mul(A).sum(dim=1).sum(dim=1).sqrt()
        Y = A.div(normA.view(batchSize, 1, 1).expand_as(A))
        K = torch.eye(dim, dim).view(1, dim, dim).repeat(batchSize, 1, 1)
        Z = torch.eye(dim, dim).view(1, dim, dim).repeat(batchSize, 1, 1)
        K = K.type(dtype)
        Z = Z.type(dtype)
        for i in range(numIters):
            T = 0.5 * (3.0 * K - Z.bmm(Y))
            Y = Y.bmm(T)
            Z = T.bmm(Z)
        sA = Y * torch.sqrt(normA).view(batchSize, 1, 1).expand_as(A)
    return sA


def calculate_frechet_inception_distance(
    acts: Union[torch.FloatTensor, np.ndarray],
    mu: np.ndarray,
    sigma: np.ndarray,
    use_torch: bool = False,
    eps: float = 1e-6,
    device: torch.device = torch.device('cuda:0'),
) -> float: # noqa
    if use_torch:
        m1 = torch.mean(acts, axis=0)
        s1 = torch_cov(acts, rowvar=False)
        mu = torch.tensor(mu).to(m1.dtype).to(device)
        sigma = torch.tensor(sigma).to(s1.dtype).to(device)
    else:
        m1 = np.mean(acts, axis=0)
        s1 = np.cov(acts, rowvar=False)
    return calculate_frechet_distance(m1, s1, mu, sigma, use_torch, eps)


def calculate_frechet_distance(
    mu1: Union[torch.FloatTensor, np.ndarray],
    sigma1: Union[torch.FloatTensor, np.ndarray],
    mu2: Union[torch.FloatTensor, np.ndarray],
    sigma2: Union[torch.FloatTensor, np.ndarray],
    use_torch: bool = False,
    eps: float = 1e-6,
) -> float:
    """Calculate Frechet Distance.

    Args:
        mu1: The sample mean over activations for a set of samples.
        sigma1: The covariance matrix over activations for a set of samples.
        mu2: The sample mean over activations for another set of samples.
        sigma2: The covariance matrix over activations for another set of
                samples.
        use_torch: When True, use torch to calculate FID. Otherwise, use numpy.
        eps: prevent covmean from being singular matrix

    Returns:
        The Frechet Distance.
    """
    if use_torch:
        assert mu1.shape == mu2.shape, \
            'Training and test mean vectors have different lengths'
        assert sigma1.shape == sigma2.shape, \
            'Training and test covariances have different dimensions'

        diff = mu1 - mu2
        # Run 50 itrs of newton-schulz to get the matrix sqrt of
        # sigma1 dot sigma2
        covmean = sqrt_newton_schulz(sigma1.mm(sigma2).unsqueeze(0), 50)
        if torch.any(torch.isnan(covmean)):
            return float('nan')
        covmean = covmean.squeeze()
        out = (diff.dot(diff) +                         # noqa: W504
               torch.trace(sigma1) +                    # noqa: W504
               torch.trace(sigma2) -                    # noqa: W504
               2 * torch.trace(covmean)).cpu().item()
    else:
        mu1 = np.atleast_1d(mu1)
        mu2 = np.atleast_1d(mu2)

        sigma1 = np.atleast_2d(sigma1)
        sigma2 = np.atleast_2d(sigma2)

        assert mu1.shape == mu2.shape, \
            'Training and test mean vectors have different lengths'
        assert sigma1.shape == sigma2.shape, \
            'Training and test covariances have different dimensions'

        diff = mu1 - mu2

        # Product might be almost singular
        covmean, _ = linalg.sqrtm(sigma1.dot(sigma2), disp=False)
        if not np.isfinite(covmean).all():
            msg = ('fid calculation produces singular product; '
                   'adding %s to diagonal of cov estimates') % eps
            print(msg)
            offset = np.eye(sigma1.shape[0]) * eps
            covmean = linalg.sqrtm((sigma1 + offset).dot(sigma2 + offset))

        # Numerical error might give slight imaginary component
        if np.iscomplexobj(covmean):
            if not np.allclose(np.diagonal(covmean).imag, 0, atol=1e-3):
                m = np.max(np.abs(covmean.imag))
                raise ValueError('Imaginary component {}'.format(m))
            covmean = covmean.real

        tr_covmean = np.trace(covmean)

        out = (diff.dot(diff) +         # noqa: W504
               np.trace(sigma1) +       # noqa: W504
               np.trace(sigma2) -       # noqa: W504
               2 * tr_covmean).item()
    return out


def calculate_inception_score(
    probs: Union[torch.FloatTensor, np.ndarray],
    splits: int = 10,
    use_torch: bool = False,
) -> Tuple[float, float]: # noqa
    # Inception Score
    scores = []
    for i in range(splits):
        part = probs[
            (i * probs.shape[0] // splits):
            ((i + 1) * probs.shape[0] // splits), :]
        if use_torch:
            kl = part * (
                torch.log(part) -                                   # noqa: W504
                torch.log(torch.unsqueeze(torch.mean(part, 0), 0)))
            kl = torch.mean(torch.sum(kl, 1))
            scores.append(torch.exp(kl))
        else:
            kl = part * (
                np.log(part) -                                      # noqa: W504
                np.log(np.expand_dims(np.mean(part, 0), 0)))
            kl = np.mean(np.sum(kl, 1))
            scores.append(np.exp(kl))
    if use_torch:
        scores = torch.stack(scores)
        inception_score = torch.mean(scores).cpu().item()
        std = torch.std(scores).cpu().item()
    else:
        inception_score, std = (np.mean(scores).item(), np.std(scores).item())
    del probs, scores
    return inception_score, std



"""Calculate the FID and Inception Score of images in a directory."""

import argparse
import os
import tempfile

import torch
from torch.utils.data import DataLoader, SequentialSampler
from torch.utils.data.distributed import DistributedSampler

from .districuted import init, world_size, print0
from .utils import ImageDataset, get_inception_score_and_fid


def calc(args):
    """Calculate the FID and Inception Score of images in a directory."""
    dataset = ImageDataset(root=args.path, num_images=args.num_images)
    if world_size() > 1:
        sampler = DistributedSampler(dataset, shuffle=False)
    else:
        sampler = SequentialSampler(dataset)
    loader = DataLoader(
        dataset,
        batch_size=args.batch_size,
        sampler=sampler,
        num_workers=args.num_workers)
    (IS, IS_std), FID = get_inception_score_and_fid(
        loader,
        args.fid_ref,
        use_torch=args.use_torch,
        verbose=True)
    print0(IS, IS_std, FID)


def calc_init(init_method, world_size, rank, args):
    """Initialize the distributed environment and calculate the FID and Inception Score of images in a directory."""
    init(init_method, world_size, rank)
    calc(args)


def main():
    """Parse command-line arguments and calculate the FID and Inception Score of images in a directory."""
    parser = argparse.ArgumentParser(
        description="A command-line tool to calculate Frechet Inception Distance (FID) between generated and reference images.",
        epilog="Example: CUDA_VISIBLE_DEVICES=0,1 python -m pytorch_image_generation_metrics.calc_metrics --path cifar10/train --fid_ref cifar10.test.npz --batch 64",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--path', type=str, required=True,
                        help='Path to the directory containing generated images.')
    parser.add_argument('--fid_ref', type=str, required=True,
                        help='Path to precalculated reference statistics file.')
    parser.add_argument("--batch_size", type=int, default=50,
                        help="Batch size for processing images.")
    parser.add_argument("--num_images", type=int, default=None,
                        help="Number of images to use for calculating FID. If not specified, all images in the directory will be used.")
    parser.add_argument('--use_torch', action='store_true',
                        help='Use PyTorch for matrix operations.')
    parser.add_argument("--num_workers", type=int, default=os.cpu_count(),
                        help="Number of worker processes for data loading.")
    args = parser.parse_args()

    world_size = len(os.environ.get('CUDA_VISIBLE_DEVICES', "0").split(','))
    if world_size == 1:
        calc(args)
    else:
        with tempfile.TemporaryDirectory() as temp:
            init_method = f'file://{os.path.abspath(os.path.join(temp, ".ddp"))}'
            processes = []
            for rank in range(world_size):
                p = torch.multiprocessing.Process(
                    target=calc_init,
                    args=(init_method, world_size, rank, args))
                p.start()
                processes.append(p)
            for p in processes:
                p.join()


if __name__ == "__main__":
    main()



"""Utilities for distributed processing in PyTorch."""

import torch
import torch.distributed


def init(init_method, world_size, rank):
    """Initialize the distributed process group for multi-GPU processing.

    Args:
        init_method (str): URL specifying how to initialize the process group.
        world_size (int): Number of processes participating in the job.
        rank (int): Rank of the current process.

    Initializes the NCCL backend for distributed GPU communication and sets the current CUDA device to the given rank.
    """
    torch.distributed.init_process_group('nccl', init_method, world_size=world_size, rank=rank)
    torch.cuda.set_device(rank)
    torch.cuda.empty_cache()


def rank():
    """Return the rank of the current process in the distributed process group.

    Returns:
        int: Rank of the current process. Returns 0 if the process group is not initialized.
    """
    if torch.distributed.is_initialized():
        return torch.distributed.get_rank()
    else:
        return 0


def world_size():
    """Return the number of processes in the distributed process group.

    Returns:
        int: Number of processes. Returns 1 if the process group is not initialized.
    """
    if torch.distributed.is_initialized():
        return torch.distributed.get_world_size()
    else:
        return 1


def barrier():
    """Synchronize all processes in the distributed process group.

    Blocks until all processes have reached this function call.
    """
    if torch.distributed.is_initialized():
        torch.distributed.barrier()


def device():
    """Return the current CUDA device for the process based on its rank.

    Returns:
        torch.device: CUDA device object for the current process.
    """
    return torch.device(f'cuda:{rank()}')


def gather_shape(x: torch.Tensor, dim: int = 0):
    """Gather the shapes of tensors along a specific dimension from all processes in the distributed process group.

    Args:
        x (torch.Tensor): The tensor whose shape to gather.
        dim (int): The dimension along which to gather the shapes. Default is 0.

    Returns:
        list of torch.Size: A list of shapes from all processes.
    """
    if world_size() > 1:
        sizes_at_dim = [torch.tensor(0).to(x.device) for _ in range(world_size())]
        torch.distributed.all_gather(sizes_at_dim, torch.tensor(x.shape[dim], device=x.device))
        shapes = []
        for size in sizes_at_dim:
            shape = list(x.shape)
            shape[dim] = size.item()
            shapes.append(torch.Size(shape))
        return shapes
    else:
        return [x.shape]


def gather(x: torch.Tensor, cat_dim: int = 0):
    """Gather tensors from all processes and concatenates them along a specified dimension.

    Args:
        x (torch.Tensor): The tensor to gather.
        cat_dim (int): The dimension along which to concatenate the tensors. Default is 0.

    Returns:
        torch.Tensor: The concatenated tensor from all processes.
    """
    if world_size() > 1:
        shapes = gather_shape(x, cat_dim)
        xs = [torch.zeros(shape, device=x.device) for shape in shapes]
        torch.distributed.all_gather(xs, x)
        return torch.cat(xs, dim=cat_dim)
    else:
        return x


def print0(*args, **kwargs):
    """Print messages only from the process with rank 0."""
    if rank() == 0:
        print(*args, **kwargs)



"""Inception Model v3 for FID computation."""

from packaging import version

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision
from torchvision import models
from torch.hub import load_state_dict_from_url

# Inception weights ported to Pytorch from
# http://download.tensorflow.org/models/image/imagenet/inception-2015-12-05.tgz
FID_WEIGHTS_URL = ('https://github.com/w86763777/pytorch_image_generation_metrics/releases/'
                   'download/v0.1.0/pt_inception-2015-12-05-6726825d.pth')
TORCHVISION_VERSION = version.parse(torchvision.__version__)


class InceptionV3(nn.Module):
    """Pretrained InceptionV3 network returning feature maps."""

    # Index of default block of inception to return,
    # corresponds to output of final average pooling
    DEFAULT_BLOCK_INDEX = 3

    # Maps feature dimensionality to their output blocks indices
    BLOCK_INDEX_BY_DIM = {
        64: 0,      # First max pooling features
        192: 1,     # Second max pooling featurs
        768: 2,     # Pre-aux classifier features
        2048: 3,    # Final average pooling features
        1008: 4,    # softmax layer
    }

    def __init__(self,
                 output_blocks=[DEFAULT_BLOCK_INDEX],
                 resize_input=True,
                 normalize_input=True,
                 requires_grad=False,
                 use_fid_inception=True):
        """Build pretrained InceptionV3.

        Args:
            output_blocks : List of int
                Indices of blocks to return features of. Possible values are:
                    - 0: corresponds to output of first max pooling
                    - 1: corresponds to output of second max pooling
                    - 2: corresponds to output which is fed to aux classifier
                    - 3: corresponds to output of final average pooling
                    - 4: corresponds to output of softmax
            resize_input : bool
                If true, bilinearly resizes input to width and height 299
                before feeding input to model. As the network without fully
                connected layers is fully convolutional, it should be able to
                handle inputs of arbitrary size, so resizing might not be
                strictly needed
            normalize_input : bool
                If true, scales the input from range (0, 1) to the range the
                pretrained Inception network expects, namely (-1, 1)
            requires_grad : bool
                If true, parameters of the model require gradients. Possibly
                useful for finetuning the network
            use_fid_inception : bool
                If true, uses the pretrained Inception model used in
                Tensorflow's FID implementation. If false, uses the pretrained
                Inception model available in torchvision. The FID Inception
                model has different weights and a slightly different structure
                from torchvision's Inception model. If you want to compute FID
                scores, you are strongly advised to set this parameter to true
                to get comparable results.
        """
        super(InceptionV3, self).__init__()

        self.resize_input = resize_input
        self.normalize_input = normalize_input
        self.output_blocks = output_blocks
        self.last_needed_block = max(output_blocks)

        # assert self.last_needed_block <= 3, \
        #     'Last possible output block index is 3'

        self.blocks = nn.ModuleList()

        if use_fid_inception:
            inception = fid_inception_v3()
        else:
            if TORCHVISION_VERSION < version.parse("0.13.0"):
                inception = models.inception_v3(
                    pretrained=True,
                    init_weights=False)
            else:
                inception = models.inception_v3(
                    weights=models.Inception_V3_Weights.IMAGENET1K_V1,
                )

        # Block 0: input to maxpool1
        block0 = [
            inception.Conv2d_1a_3x3,
            inception.Conv2d_2a_3x3,
            inception.Conv2d_2b_3x3,
            nn.MaxPool2d(kernel_size=3, stride=2)
        ]
        self.blocks.append(nn.Sequential(*block0))

        # Block 1: maxpool1 to maxpool2
        if self.last_needed_block >= 1:
            block1 = [
                inception.Conv2d_3b_1x1,
                inception.Conv2d_4a_3x3,
                nn.MaxPool2d(kernel_size=3, stride=2)
            ]
            self.blocks.append(nn.Sequential(*block1))

        # Block 2: maxpool2 to aux classifier
        if self.last_needed_block >= 2:
            block2 = [
                inception.Mixed_5b,
                inception.Mixed_5c,
                inception.Mixed_5d,
                inception.Mixed_6a,
                inception.Mixed_6b,
                inception.Mixed_6c,
                inception.Mixed_6d,
                inception.Mixed_6e,
            ]
            self.blocks.append(nn.Sequential(*block2))

        # Block 3: aux classifier to final avgpool
        if self.last_needed_block >= 3:
            block3 = [
                inception.Mixed_7a,
                inception.Mixed_7b,
                inception.Mixed_7c,
                nn.AdaptiveAvgPool2d(output_size=(1, 1))
            ]
            self.blocks.append(nn.Sequential(*block3))

        if self.last_needed_block >= 4:
            inception.fc.bias = None
            self.blocks.append(inception.fc)

        for param in self.parameters():
            param.requires_grad = requires_grad

    def forward(self, x):
        """Get Inception feature maps.

        Args:
            x : torch.FloatTensor,Input tensor of shape [B x 3 x H x W]. If
                `normalize_input` is True, values are expected to be in range
                [0, 1]; Otherwise, values are expected to be in range [-1, 1].

        Returns:
            List of torch.FloatTensor, corresponding to the selected output
            block, sorted ascending by index
        """
        outputs = [None for _ in range(len(self.output_blocks))]

        if self.resize_input:
            x = F.interpolate(x,
                              size=(299, 299),
                              mode='bilinear',
                              align_corners=False)

        if self.normalize_input:
            x = 2 * x - 1  # Scale from range (0, 1) to range (-1, 1)

        for idx, block in enumerate(self.blocks):
            if idx < 4:
                x = block(x)
            else:
                x = F.dropout(x, training=self.training)    # N x 2048 x 1 x 1
                x = torch.flatten(x, start_dim=1)           # N x 2048
                x = block(x)                                # N x 1000
                x = F.softmax(x, dim=1)

            if idx in self.output_blocks:
                order = self.output_blocks.index(idx)
                outputs[order] = x

            if idx == self.last_needed_block:
                break

        return outputs


def fid_inception_v3():
    """Build pretrained Inception model for FID computation.

    The Inception model for FID computation uses a different set of weights
    and has a slightly different structure than torchvision's Inception.

    This method first constructs torchvision's Inception and then patches the
    necessary parts that are different in the FID Inception model.
    """
    if TORCHVISION_VERSION < version.parse("0.13.0"):
        inception = models.inception_v3(
            pretrained=False,
            aux_logits=False,
            num_classes=1008,
            init_weights=False,
        )
    else:
        inception = models.inception_v3(
            weights=None,
            aux_logits=False,
            num_classes=1008,
            init_weights=False,
        )
    inception.Mixed_5b = FIDInceptionA(192, pool_features=32)
    inception.Mixed_5c = FIDInceptionA(256, pool_features=64)
    inception.Mixed_5d = FIDInceptionA(288, pool_features=64)
    inception.Mixed_6b = FIDInceptionC(768, channels_7x7=128)
    inception.Mixed_6c = FIDInceptionC(768, channels_7x7=160)
    inception.Mixed_6d = FIDInceptionC(768, channels_7x7=160)
    inception.Mixed_6e = FIDInceptionC(768, channels_7x7=192)
    inception.Mixed_7b = FIDInceptionE_1(1280)
    inception.Mixed_7c = FIDInceptionE_2(2048)

    state_dict = load_state_dict_from_url(FID_WEIGHTS_URL, progress=True)
    inception.load_state_dict(state_dict)
    return inception


class FIDInceptionA(models.inception.InceptionA):
    """InceptionA block patched for FID computation."""

    def __init__(self, in_channels, pool_features): # noqa
        super(FIDInceptionA, self).__init__(in_channels, pool_features)

    def forward(self, x):                           # noqa
        branch1x1 = self.branch1x1(x)

        branch5x5 = self.branch5x5_1(x)
        branch5x5 = self.branch5x5_2(branch5x5)

        branch3x3dbl = self.branch3x3dbl_1(x)
        branch3x3dbl = self.branch3x3dbl_2(branch3x3dbl)
        branch3x3dbl = self.branch3x3dbl_3(branch3x3dbl)

        # Patch: Tensorflow's average pool does not use the padded zero's in
        # its average calculation
        branch_pool = F.avg_pool2d(x, kernel_size=3, stride=1, padding=1,
                                   count_include_pad=False)
        branch_pool = self.branch_pool(branch_pool)

        outputs = [branch1x1, branch5x5, branch3x3dbl, branch_pool]
        return torch.cat(outputs, 1)


class FIDInceptionC(models.inception.InceptionC):
    """InceptionC block patched for FID computation."""

    def __init__(self, in_channels, channels_7x7):  # noqa
        super(FIDInceptionC, self).__init__(in_channels, channels_7x7)

    def forward(self, x):                           # noqa
        branch1x1 = self.branch1x1(x)

        branch7x7 = self.branch7x7_1(x)
        branch7x7 = self.branch7x7_2(branch7x7)
        branch7x7 = self.branch7x7_3(branch7x7)

        branch7x7dbl = self.branch7x7dbl_1(x)
        branch7x7dbl = self.branch7x7dbl_2(branch7x7dbl)
        branch7x7dbl = self.branch7x7dbl_3(branch7x7dbl)
        branch7x7dbl = self.branch7x7dbl_4(branch7x7dbl)
        branch7x7dbl = self.branch7x7dbl_5(branch7x7dbl)

        # Patch: Tensorflow's average pool does not use the padded zero's in
        # its average calculation
        branch_pool = F.avg_pool2d(x, kernel_size=3, stride=1, padding=1,
                                   count_include_pad=False)
        branch_pool = self.branch_pool(branch_pool)

        outputs = [branch1x1, branch7x7, branch7x7dbl, branch_pool]
        return torch.cat(outputs, 1)


class FIDInceptionE_1(models.inception.InceptionE):
    """First InceptionE block patched for FID computation."""

    def __init__(self, in_channels):    # noqa
        super(FIDInceptionE_1, self).__init__(in_channels)

    def forward(self, x):               # noqa
        branch1x1 = self.branch1x1(x)

        branch3x3 = self.branch3x3_1(x)
        branch3x3 = [
            self.branch3x3_2a(branch3x3),
            self.branch3x3_2b(branch3x3),
        ]
        branch3x3 = torch.cat(branch3x3, 1)

        branch3x3dbl = self.branch3x3dbl_1(x)
        branch3x3dbl = self.branch3x3dbl_2(branch3x3dbl)
        branch3x3dbl = [
            self.branch3x3dbl_3a(branch3x3dbl),
            self.branch3x3dbl_3b(branch3x3dbl),
        ]
        branch3x3dbl = torch.cat(branch3x3dbl, 1)

        # Patch: Tensorflow's average pool does not use the padded zero's in
        # its average calculation
        branch_pool = F.avg_pool2d(x, kernel_size=3, stride=1, padding=1,
                                   count_include_pad=False)
        branch_pool = self.branch_pool(branch_pool)

        outputs = [branch1x1, branch3x3, branch3x3dbl, branch_pool]
        return torch.cat(outputs, 1)


class FIDInceptionE_2(models.inception.InceptionE):
    """Second InceptionE block patched for FID computation."""

    def __init__(self, in_channels):    # noqa
        super(FIDInceptionE_2, self).__init__(in_channels)

    def forward(self, x):               # noqa
        branch1x1 = self.branch1x1(x)

        branch3x3 = self.branch3x3_1(x)
        branch3x3 = [
            self.branch3x3_2a(branch3x3),
            self.branch3x3_2b(branch3x3),
        ]
        branch3x3 = torch.cat(branch3x3, 1)

        branch3x3dbl = self.branch3x3dbl_1(x)
        branch3x3dbl = self.branch3x3dbl_2(branch3x3dbl)
        branch3x3dbl = [
            self.branch3x3dbl_3a(branch3x3dbl),
            self.branch3x3dbl_3b(branch3x3dbl),
        ]
        branch3x3dbl = torch.cat(branch3x3dbl, 1)

        # Patch: The FID Inception model uses max pooling instead of average
        # pooling. This is likely an error in this specific Inception
        # implementation, as other Inception models use average pooling here
        # (which matches the description in the paper).
        branch_pool = F.max_pool2d(x, kernel_size=3, stride=1, padding=1)
        branch_pool = self.branch_pool(branch_pool)

        outputs = [branch1x1, branch3x3, branch3x3dbl, branch_pool]
        return torch.cat(outputs, 1)



"""Export public API."""

from .version import __version__

from pytorch_image_generation_metrics.utils import (
    ImageDataset,
    get_inception_score,
    get_inception_score_from_directory,
    get_fid,
    get_fid_from_directory,
    get_inception_score_and_fid,
    get_inception_score_and_fid_from_directory)

__all__ = [
    ImageDataset,
    get_inception_score,
    get_inception_score_from_directory,
    get_fid,
    get_fid_from_directory,
    get_inception_score_and_fid,
    get_inception_score_and_fid_from_directory,
    __version__,
]



"""The public API of pytorch_image_generation_metrics."""

import os
from typing import List, Union, Tuple, Optional
from glob import glob

import numpy as np
import torch
from PIL import Image
from torch.utils.data import Dataset, DataLoader, SequentialSampler
from torch.utils.data.distributed import DistributedSampler
from torchvision.transforms import Compose, Resize, ToTensor
from torchvision.transforms.functional import to_tensor

from .districuted import rank, world_size
from .core import (
    get_inception_feature,
    calculate_inception_score,
    calculate_frechet_inception_distance,
    torch_cov)


class ImageDataset(Dataset):
    """An simple image dataset for calculating inception score and FID."""

    def __init__(self, root, exts=['png', 'jpg', 'JPEG'], transform=None,
                 num_images=None):
        """Construct an image dataset.

        Args:
            root: Path to the image directory. This directory will be
                  recursively searched.
            exts: List of extensions to search for.
            transform: A torchvision transform to apply to the images. If
                       None, the images will be converted to tensors.
            num_images: The number of images to load. If None, all images
                        will be loaded.
        """
        self.paths = []
        self.transform = transform
        for ext in exts:
            self.paths.extend(
                list(glob(
                    os.path.join(root, '**/*.%s' % ext), recursive=True)))
        self.paths = self.paths[:num_images]

    def __len__(self):              # noqa
        return len(self.paths)

    def __getitem__(self, idx):     # noqa
        image = Image.open(self.paths[idx])
        image = image.convert('RGB')        # fix ImageNet grayscale images
        if self.transform is not None:
            image = self.transform(image)
        else:
            image = to_tensor(image)
        return image


def get_inception_score_and_fid(
    images: Union[torch.FloatTensor, DataLoader],
    fid_ref: str,
    splits: int = 10,
    use_torch: bool = False,
    **kwargs,
) -> Tuple[Tuple[float, float], float]:
    """Calculate Inception Score and FID.

    For each image, only a forward propagation is required to
    calculating features for FID and Inception Score.

    Args:
        images: List of tensor or torch.utils.data.Dataloader. The return image
                must be float tensor of range [0, 1].
        fid_ref: Path to pre-calculated statistic.
        splits: The number of bins of Inception Score.
        use_torch: When True, use torch to calculate FID. Otherwise, use numpy.
        **kwargs: The arguments passed to
                  `pytorch_image_generation_metrics.core.get_inception_feature`.
    Returns:
        inception_score: float tuple, (mean, std)
        fid: float
    """
    acts, probs = get_inception_feature(
        images, dims=[2048, 1008], use_torch=use_torch, **kwargs)

    if rank() != 0:
        return (None, None), None

    # Inception Score
    inception_score, std = calculate_inception_score(probs, splits, use_torch)

    # Frechet Inception Distance
    f = np.load(fid_ref, allow_pickle=True)
    if isinstance(f, np.ndarray):
        mu, sigma = f.item()['mu'][:], f.item()['sigma'][:]
    else:
        mu, sigma = f['mu'][:], f['sigma'][:]
        f.close()
    fid = calculate_frechet_inception_distance(acts, mu, sigma, use_torch)

    return (inception_score, std), fid


def get_inception_score_and_fid_from_directory(
    path: str,
    fid_ref: str,
    exts: List[str] = ['png', 'jpg'],
    batch_size: int = 50,
    splits: int = 10,
    use_torch: bool = False,
    **kwargs
) -> Tuple[Tuple[float, float], float]:
    """Calculate Inception Score and FID of images in a directory.

    Args:
        path: Path to the image directory. This function will recursively find
              images in all subfolders.
        fid_ref: Path to pre-calculated statistic.
        exts: List of extensions to search for.
        batch_size: Batch size of DataLoader.
        splits: The number of bins of Inception Score.
        use_torch: When True, use torch to calculate FID. Otherwise, use numpy.
        **kwargs: The arguments passed to
                  `pytorch_image_generation_metrics.core.get_inception_feature`.

    Returns:
        Inception Score: float tuple, mean and std
        FID: float
    """
    return get_inception_score_and_fid(
        images=DataLoader(ImageDataset(path, exts), batch_size=batch_size),
        fid_ref=fid_ref,
        splits=splits,
        use_torch=use_torch, **kwargs)


def get_fid(
    images: Union[torch.FloatTensor, DataLoader],
    fid_ref: str,
    use_torch: bool = False,
    **kwargs,
) -> float:
    """Calculate Frechet Inception Distance.

    Args:
        images: List of tensor or torch.utils.data.Dataloader. The return image
                must be float tensor of range [0, 1].
        fid_ref: Path to pre-calculated statistic.
        use_torch: When True, use torch to calculate FID. Otherwise, use numpy.
        **kwargs: The arguments passed to
                  `pytorch_image_generation_metrics.core.get_inception_feature`.

    Returns:
        FID
    """
    acts, = get_inception_feature(
        images, dims=[2048], use_torch=use_torch, **kwargs)

    if rank() != 0:
        return None

    # Frechet Inception Distance
    f = np.load(fid_ref, allow_pickle=True)
    if isinstance(f, np.ndarray):
        mu, sigma = f.item()['mu'][:], f.item()['sigma'][:]
    else:
        mu, sigma = f['mu'][:], f['sigma'][:]
        f.close()
    fid = calculate_frechet_inception_distance(acts, mu, sigma, use_torch)

    return fid


def get_fid_from_directory(
    path: str,
    fid_ref: str,
    exts: List[str] = ['png', 'jpg'],
    batch_size: int = 50,
    use_torch: bool = False,
    **kwargs
) -> float:
    """Calculate Frechet Inception Distance of images in a directory.

    Args:
        path: Path to the image directory. This function will recursively find
              images in all subfolders.
        fid_ref: Path to pre-calculated statistic.
        exts: List of extensions to search for.
        use_torch: When True, use torch to calculate FID. Otherwise, use numpy.
        **kwargs: The arguments passed to
                  `pytorch_image_generation_metrics.core.get_inception_feature`.

    Returns:
        FID
    """
    return get_fid(
        images=DataLoader(ImageDataset(path, exts), batch_size=batch_size),
        fid_ref=fid_ref,
        use_torch=use_torch,
        **kwargs)


def get_inception_score(
    images: Union[torch.FloatTensor, DataLoader],
    splits: int = 10,
    use_torch: bool = False,
    **kwargs,
) -> Tuple[float, float]:
    """Calculate Inception Score.

    Args:
        images: List of tensor or torch.utils.data.Dataloader. The return image
                must be float tensor of range [0, 1].
        splits: The number of bins of Inception Score.
        use_torch: When True, use torch to calculate FID. Otherwise, use numpy.
        **kwargs: The arguments passed to
                  `pytorch_image_generation_metrics.core.get_inception_feature`.

    Returns:
        Inception Score
    """
    probs, = get_inception_feature(
        images, dims=[1008], use_torch=use_torch, **kwargs)
    if rank() != 0:
        return (None, None)
    inception_score, std = calculate_inception_score(probs, splits, use_torch)
    return (inception_score, std)


def get_inception_score_from_directory(
    path: str,
    splits: int = 10,
    exts: List[str] = ['png', 'jpg'],
    batch_size: int = 50,
    use_torch: bool = False,
    **kwargs
) -> Tuple[Tuple[float, float], float]:
    """Calculate Frechet Inception Distance of images in a directory.

    Args:
        path: Path to the image directory. This function will recursively find
              images in all subfolders.
        splits: The number of bins of Inception Score.
        exts: List of extensions to search for.
        batch_size: Batch size of DataLoader.
        use_torch: When True, use torch to calculate FID. Otherwise, use numpy.
        **kwargs: The arguments passed to
                  `pytorch_image_generation_metrics.core.get_inception_feature`.


    Returns:
        FID: float
    """
    return get_inception_score(
        images=DataLoader(ImageDataset(path, exts), batch_size=batch_size),
        splits=splits,
        use_torch=use_torch,
        **kwargs)


def calc_fid_ref(
    input_path: str,
    output_path: str = None,
    batch_size: int = 50,
    img_size: Optional[int] = None,
    use_torch: bool = False,
    num_workers: int = os.cpu_count(),
    verbose: bool = True,
) -> None:
    """Calculate the FID statistics and save them to a file.

    Args:
        input_path (str): Path to the image directory. This function will
                          recursively find images in all subfolders.
        output_path (str): Path to the output file. Use None to disable.
        batch_size (int): Batch size. Defaults to 50.
        img_size (int): Image size. If None, use the original image size.
        num_workers (int): Number of dataloader workers. Default:
                           os.cpu_count().
    """
    if img_size is not None:
        transform = Compose([Resize([img_size, img_size]), ToTensor()])
    else:
        transform = ToTensor()

    dataset = ImageDataset(root=input_path, transform=transform)
    if world_size() > 1:
        sampler = DistributedSampler(dataset, shuffle=False)
    else:
        sampler = SequentialSampler(dataset)
    loader = DataLoader(
        dataset,
        batch_size=batch_size,
        sampler=sampler,
        num_workers=num_workers)
    acts, = get_inception_feature(
        loader, dims=[2048], use_torch=use_torch, verbose=verbose)

    if rank() != 0:
        return

    if use_torch:
        mu = torch.mean(acts, dim=0).cpu().numpy()
        sigma = torch_cov(acts, rowvar=False).cpu().numpy()
    else:
        mu = np.mean(acts, axis=0)
        sigma = np.cov(acts, rowvar=False)

    if output_path is not None:
        if os.path.dirname(output_path) != "":
            os.makedirs(os.path.dirname(output_path), exist_ok=True)
        np.savez_compressed(output_path, mu=mu, sigma=sigma)

    return mu, sigma



"""Install the package."""

import os
import setuptools


def read(rel_path):
    """Read a file."""
    base_path = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(base_path, rel_path), 'r') as f:
        return f.read()


if __name__ == '__main__':
    # get __version__
    with open('./pytorch_image_generation_metrics/version.py') as f:
        exec(f.read())

    setuptools.setup(
        name='pytorch_image_generation_metrics',
        version=__version__,    # noqa: F821
        author='Yi-Lun Wu',
        author_email='w86763777@gmail.com',
        description=('Package for calculating image generation metrics using Pytorch'),
        long_description=read('README.md'),
        long_description_content_type='text/markdown',
        url='https://github.com/w86763777/pytorch_image_generation_metrics',
        packages=setuptools.find_packages(include=['pytorch_image_generation_metrics']),
        keywords=[
            'PyTorch',
            'Inception Score',
            'IS',
            'Frechet Inception Distance',
            'FID'],
        classifiers=[
            'Programming Language :: Python :: 3',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
        ],
        python_requires='>=3.6',
        install_requires=[
            "packaging",
            "tqdm",
            "scipy",
            "torch>=1.8.2",
            "torchvision>=0.9.2",
        ],
    )



"""Calculate statistics for FID and save them to a file."""

import argparse
import os
import tempfile

import torch

from .districuted import init
from .utils import calc_fid_ref


def calc(args):
    """Calculate statistics for FID and save them to a file."""
    calc_fid_ref(
        args.path,
        args.output,
        args.batch_size,
        args.img_size,
        args.use_torch,
        args.num_workers)


def calc_init(init_method, world_size, rank, args):
    """Initialize the distributed environment and calculate statistics for FID and save them to a file."""
    init(init_method, world_size, rank)
    calc(args)


def main():
    """Parse command-line arguments and calculate statistics for FID and save them to a file."""
    parser = argparse.ArgumentParser(
        description="A command-line tool to compute Frechet Inception Distance (FID) statistics.",
        epilog="Example: CUDA_VISIBLE_DEVICES=0,1 python -m pytorch_image_generation_metrics.fid_ref --path cifar10/train --output cifar10.test.npz --batch_size 64",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--path", type=str, required=True,
                        help='Path to the directory containing images (including subfolders).')
    parser.add_argument("--output", type=str, required=True,
                        help="Output file path for saving the computed statistics.")
    parser.add_argument("--batch_size", type=int, default=50,
                        help="Batch size for processing images.")
    parser.add_argument("--img_size", type=int, default=None,
                        help="Resize images to this specified size (if provided).")
    parser.add_argument('--use_torch', action='store_true',
                        help='Use PyTorch for matrix operations.')
    parser.add_argument("--num_workers", type=int, default=os.cpu_count(),
                        help="Number of worker processes for data loading.")
    args = parser.parse_args()

    world_size = len(os.environ.get('CUDA_VISIBLE_DEVICES', "0").split(','))
    if world_size == 1:
        calc(args)
    else:
        with tempfile.TemporaryDirectory() as temp:
            init_method = f'file://{os.path.abspath(os.path.join(temp, ".ddp"))}'
            processes = []
            for rank in range(world_size):
                p = torch.multiprocessing.Process(
                    target=calc_init,
                    args=(init_method, world_size, rank, args))
                p.start()
                processes.append(p)
            for p in processes:
                p.join()


if __name__ == '__main__':
    main()
